//
// CPU
//
//	ؿݥ󥿥١Ǥμ
//
//		(c)2008 Chihiro.SAKAMOTO HyperWorks
//
#include <exception>
#include "vm.h"

namespace vm {

	// 0㳰
	class devide_by_zero: public std::exception {
	  public:
		const char *what() const throw()
		{
			return "devide by zero";
		}
	} ;

	void (vcpu::*vcpu::cmd_[])() = {
		#define	VM_EXECTABLE
		#include "vm_code.h"
		#undef	VM_EXECTABLE
	} ;

	// ¹
	int vcpu::run()
	{
		command_ = data_.command_;						// ץǼ
		command_size_ = data_.command_size_;			// ץ礭

		global_value.resize(data_.value_size_);			// ѿơ֥
		command_ptr_ = command_;						// ץ५󥿡

		active = true;									// HaltƤʤ

		try {
			while (active) {							// Haltޤǥ롼
				int op = *command_ptr_++;
				(this->*cmd_[op])();
			}
		}
		catch (const std::exception &e) {
			std::cerr << "㳰ȯ" << e.what() << "" << std::endl;
			return -1;
		}
		return 0;
	}

	// Push
	void vcpu::PushConst()
	{
		push(value());
	}

	// ѿPush
	void vcpu::PushValue()
	{
		push(global_value[value()]);
	}

	// ѿPop
	void vcpu::PopValue()
	{
		global_value[value()] = top(); pop();
	}

	// PopʥåȥåפΤƤ
	void vcpu::OpPop()
	{
		pop();
	}

	// ñޥʥ
	void vcpu::OpNeg()
	{
		top() = -top();
	}

	// ==
	void vcpu::OpEq()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs == rhs);
	}

	// !=
	void vcpu::OpNe()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs != rhs);
	}

	// >
	void vcpu::OpGt()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs > rhs);
	}

	// >=
	void vcpu::OpGe()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs >= rhs);
	}

	// <
	void vcpu::OpLt()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs < rhs);
	}

	// <=
	void vcpu::OpLe()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs <= rhs);
	}

	// +
	void vcpu::OpAdd()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs + rhs);
	}

	// -
	void vcpu::OpSub()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs - rhs);
	}

	// *
	void vcpu::OpMul()
	{
		int rhs = top(); pop();
		int lhs = top(); pop();
		push(lhs * rhs);
	}

	// /
	void vcpu::OpDiv()
	{
		int rhs = top(); pop();
		if (rhs == 0)
			throw devide_by_zero();
		int lhs = top(); pop();
		push(lhs / rhs);
	}

	// %
	void vcpu::OpMod()
	{
		int rhs = top(); pop();
		if (rhs == 0)
			throw devide_by_zero();
		int lhs = top(); pop();
		push(lhs % rhs);
	}

	// ̵兩
	void vcpu::OpJmp()
	{
		jmp(value());
	}

	// λ
	void vcpu::OpJmpC()
	{
		int addr = value();
		int cond = top(); pop();
		if (cond)
			jmp(addr);
	}

	// λ
	void vcpu::OpJmpNC()
	{
		int addr = value();
		int cond = top(); pop();
		if (!cond)
			jmp(addr);
	}

	// CPUץ
	void vcpu::OpHalt()
	{
		active = false;
	}

	void vcpu::OpRand()
	{
		int range = top(); pop();
		int value = (range <= 0)? 0: (rand() % range);
		push(value);
	}

	void vcpu::OpPrint()
	{
		int count = value();
		while (count--) {
			std::cout << top();
			pop();
			if (count)
				std::cout << ", ";
		}
		std::cout << std::endl;
	}
}
