//
// lZp[T[
//
//	Chihiro.SAKAMOTO
//
#include <climits>
#include <cstdio>
#include "parser.h"

cparser::cparser()
	: token_type(OP_EOF)
{
}

cparser::~cparser()
{
}

// \
//
//	statement ::= <value> "=" <expr>
//				| "print" <expr>
//				| "list"
//
bool cparser::parse(calc_driver *driver, const std::string &f)
{
	if ((fp = std::fopen(f.c_str(), "r")) == NULL)
		return false;

	int token;
	std::string str;
	while ((token = get_token(str)) != OP_EOF) {
		if (token == OP_LIST) {				// listR}h
			driver->list();
		}
		else if (token == OP_PRINT) {		// printR}h
			cnode *node = add_sub_expr();
			driver->print(node);
			delete node;
		}
		else if (token == OP_VALUE) {		// 
			std::string equ;
			if (get_token(equ) != OP_EQU)
				return false;
			cnode *node = add_sub_expr();
			driver->assign(str, node);
			delete node;
		}
		else if (token != OP_NEWLINE) {		// sȊO̓G[
			return false;
		}

		token = get_token(str);
		if (token != OP_NEWLINE)			// s\nǂݎ̂Ă
			unget_token(token, str);
	}

	std::fclose(fp);

	return true;
}

// {|̏
//	expr ::= <muldiv> (("+" | "-") <muldiv>)*

cnode *cparser::add_sub_expr()
{
	cnode *left = mul_div_expr();
	if  (left == NULL)
		return NULL;

	for (;;) {
		std::string str;
		int token = get_token(str);
		if (token != OP_PLUS && token != OP_MINUS) {
			unget_token(token, str);
			break;
		}

		cnode *right = mul_div_expr();
		if (right == NULL) {
			delete left;
			return NULL;
		}
		left = new cnode(token, left, right);
	}
	return left;
}

// ^̏
//	muldiv ::= <term> (("*" | "/") <term>)*

cnode *cparser::mul_div_expr()
{
	cnode *left = term();
	if  (left == NULL)
		return NULL;

	for (;;) {
		std::string str;
		int token = get_token(str);
		if (token != OP_TIMES && token != OP_DIVIDE) {
			unget_token(token, str);
			break;
		}

		cnode *right = term();
		if (right == NULL) {
			delete left;
			return NULL;
		}
		left = new cnode(token, left, right);
	}
	return left;
}

// I[
//	term ::= "-" <term> | "(" <expr> ")" | <number> | <value>

cnode *cparser::term()
{
	std::string str;
	int token = get_token(str);

	if (token == OP_MINUS) {
		return new cnode(OP_NEG, term());
	}
	if (token == OP_OPAR) {
		cnode *node = add_sub_expr();
		if (node == NULL)
			return NULL;

		token = get_token(str);
		if (token != OP_CPAR) {
			delete node;
			unget_token(token, str);
			return NULL;
		}
		return node;
	}
	if (token == OP_CONST) {
		return new cnode(OP_CONST, atoi(str.c_str()));
	}
	if (token == OP_VALUE) {
		return new cnode(OP_VALUE, new std::string(str));
	}
	return NULL;
}

// P[̕

inline bool _isspace(int c)
{
	return c == ' ' || c == '\t';
}

inline bool _isnum(int c)
{
	return c >= '0' && c <= '9';
}

inline bool _isalpha(int c)
{
	return c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z' || c == '_';
}

// 
//
// g[N					g[N
//
// '+'						OP_PLUS
// '-'						OP_MINUS
// '*'						OP_TIMES
// '/'						OP_DIVIDE
// '('						OP_OPAR
// ')'						OP_CPAR
// '='						OP_EQU
// '\n'						OP_NEWLINE
// EOF						OP_EOF
//
// "print"					OP_PRINT
// "list"					OP_LIST
// [0-9]+					OP_CONST
// [a-zA-Z_][a-zA-Z0-9_]*	OP_VALUE
//
int cparser::_get_token(std::string &str)
{
	int c;

	// s󔒗ނǂݔ΂B
	do {
		c = std::fgetc(fp);
	} while (_isspace(c)) ;

	// str͍ėpĂ邩Ȃ̂ŁANAĂ
	str.clear();

	// L̏ꍇ́APłPg[NȂ̂ŁAʂԂB
	switch (c) {
	  case '+':		return OP_PLUS;
	  case '-':		return OP_MINUS;
	  case '*':		return OP_TIMES;
	  case '/':		return OP_DIVIDE;
	  case '(':		return OP_OPAR;
	  case ')':		return OP_CPAR;
	  case '=':		return OP_EQU;
	  case '\n':	return OP_NEWLINE;
	  case EOF:		return OP_EOF;
	}

	// Pڂ̏ꍇ́AƂăg[N؂oB
	if (_isnum(c)) {
		while (_isnum(c)) {
			str += c;
			c = std::fgetc(fp);
		}
		std::ungetc(c, fp);
		return OP_CONST;
	}

	// Pڂp̏ꍇ́AuʎqvƂăg[N؂oB
	if (_isalpha(c)) {
		while (_isalpha(c) || _isnum(c)) {
			str += c;
			c = fgetc(fp);
		}
		std::ungetc(c, fp);

		// ؂og[Nu\v̏ꍇ́AΉg[NʂԂB
		if (str == "print")
			return OP_PRINT;
		if (str == "list")
			return OP_LIST;
		return OP_VALUE;
	}

	// ̑̓̕G[ƂB
	return OP_ERROR;
}
