#include <boost/spirit.hpp>
#include <boost/spirit/phoenix.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/mem_fn.hpp>
#include <iostream>
#include <map>
#include <vector>
#include <string>
#include <cassert>
#include <algorithm>
#include <functional>
#include "node.h"

using namespace std;
using namespace boost::spirit;

typedef file_iterator<>	iterator_t;

map<string, int> values;	// ѿơ֥

// ʸμ¹Դؿ
class caction {
  public:
	virtual ~caction()
	{
	}

	virtual void action() const = 0;
} ;

typedef	boost::shared_ptr<caction> caction_t;

// ʸ
class cassign_action: public caction {
  public:
	cassign_action(const string &value, cnode_t node)
		: value_(value), node_(node)
	{
	}

	virtual void action() const;

  protected:
	string value_;
	cnode_t node_;
} ;

// ʸμ¹Դؿ
void cassign_action::action() const
{
	values[value_] = node_->expr(values);
}

// printʸ
class cprint_action: public caction {
  public:
	cprint_action(cnode_t node)
		: node_(node)
	{
	}

	virtual void action() const;

  protected:
	cnode_t node_;
} ;

// printʸμ¹Դؿ
void cprint_action::action() const
{
	std::cout << node_->expr(values) << std::endl;
}

// listʸ
class clist_action: public caction {
  public:
	clist_action()
	{
	}

	virtual void action() const;
} ;

// listʸμ¹Դؿ
struct print_value {
	void operator()(const std::pair<std::string, int> &it)
	{
		std::cout << it.first << " = " << it.second << std::endl;
	}
} ;

void clist_action::action() const
{
	for_each(values.begin(), values.end(), print_value());
}

// 顼ѡ
struct error_parser {
	typedef nil_t result_t;		// ѡη̷nil_t

	error_parser(char const *msg)
		: msg_(msg)
	{
	}

	template <typename ScannerT>
	int operator()(ScannerT const &scan, result_t &result) const
	{
		// ޤ褿-1֤
		if (scan.at_end()) {
			return -1;
		}

		// ᤷĹפ֤ȡʬåפΤ
		// ԤޤǤ򥹥󤷡ʬ򥹥å

		iterator_t b = scan.first;
		size_t length = (*(anychar_p - '\n')).parse(scan).length();
		cout << msg_ << " : " << string(b, scan.first) << endl;
		return (int)length + 1;
	}

  private:
	const char *msg_;
} ;

// 顼ѡ
typedef functor_parser<error_parser> error_p;

// ʸˡ顼ѡ
error_p syntax_error_p = error_parser("ʸˡ顼");

// ñ黻ҥΡɤ
struct unary_node_impl {
	template <typename Ty1, typename Ty2>
	struct result { typedef cnode_t type; } ;

	template <typename Ty1, typename Ty2>
	cnode_t operator()(Ty1 op, const Ty2 &left) const
	{
		return cnode_t(new cnode(op, left));
	}
} ;

// 黻ҥΡɤ
struct binary_node_impl {
	template <typename Ty1, typename Ty2, typename Ty3>
	struct result { typedef cnode_t type; } ;

	template <typename Ty1, typename Ty2, typename Ty3>
	cnode_t operator()(Ty1 op, const Ty2 &left, const Ty3 &right) const
	{
		return cnode_t(new cnode(op, left, right));
	}
} ;

// phoenixѤ̵̾ؿѤδؿ
phoenix::function<binary_node_impl> const binary_node = binary_node_impl();
phoenix::function<unary_node_impl> const unary_node = unary_node_impl();

// ʸˡ
struct calc_grammer: public grammar<calc_grammer> {
	calc_grammer(vector<caction_t> &statements)
		:statements_(statements)
	{
	}

	// ʸϿƤvector
	vector<caction_t> &statements_;

	// ѿ̾Υ
	struct ident_val: closure<ident_val, std::string> {
		member1 str;
	} ;
	// Υ
	struct number_val: closure<number_val, unsigned int> {
		member1 number;
	} ;
	// ΡɤΥ
	struct node_val: closure<node_val, cnode_t> {
		member1 node;
	} ;
	// ʸΥ
	struct assign_val: closure<assign_val, caction_t, std::string> {
		member1 action;
		member2 value;
	} ;
	// ʳΡʸΥ
	struct state_val: closure<state_val, caction_t> {
		member1 action;
	} ;

	template <typename ScannerT>
	struct definition {
		rule<ScannerT, ident_val::context_t>	identifier;
		rule<ScannerT, number_val::context_t>	number;
		rule<ScannerT, node_val::context_t>		prime;
		rule<ScannerT, node_val::context_t>		unary;
		rule<ScannerT, node_val::context_t>		mul_expr;
		rule<ScannerT, node_val::context_t>		add_expr;
		rule<ScannerT, assign_val::context_t>	assign;
		rule<ScannerT, state_val::context_t>	print;
		rule<ScannerT, state_val::context_t>	list;
		rule<ScannerT, state_val::context_t>	statement;
		rule<ScannerT>							input;
		rule<ScannerT>							ident;

		symbols<> keywords;
		dynamic_distinct_parser<ScannerT> keyword_p;

		definition(calc_grammer const& self)
			:keyword_p(alnum_p | '_')
		{
			using phoenix::arg1;
			using phoenix::arg2;
			using phoenix::var;
			using phoenix::new_;
			using phoenix::construct_;

			keywords = "print", "list";
			ident = lexeme_d[
				((alpha_p | '_') >> *(alnum_p | '_')) - (keywords >> anychar_p - (alnum_p | '_'))
			];

			identifier = ident[identifier.str = construct_<string>(arg1, arg2)];
			number = uint_p[number.number = arg1];

			prime	= identifier[prime.node = unary_node(OP_IDENT, arg1)]
					| number[prime.node = unary_node(OP_NUMBER, arg1)]
					| '(' >> add_expr[prime.node = arg1] >> ')'
					;

			unary	= prime[unary.node = arg1]
					| '-' >> prime[unary.node = unary_node(OP_NEG, arg1)]
					;

			mul_expr = unary[mul_expr.node = arg1]
					>> *('*' >> unary[mul_expr.node = binary_node(OP_TIMES, mul_expr.node, arg1)]
					   | '/' >> unary[mul_expr.node = binary_node(OP_DIVIDE, mul_expr.node, arg1)]);

			add_expr = mul_expr[add_expr.node = arg1]
					>> *('+' >> mul_expr[add_expr.node = binary_node(OP_PLUS, add_expr.node, arg1)]
					   | '-' >> mul_expr[add_expr.node = binary_node(OP_MINUS, add_expr.node, arg1)]);

			assign = identifier[assign.value = arg1] >> '='
					>> add_expr[assign.action = construct_<caction_t>(new_<cassign_action>(assign.value, arg1))] >> '\n';
			//print = lexeme_d[str_p("print") >> eps_p(anychar_p - (alnum_p | '_'))] >> add_expr[print.action = construct_<caction_t>(new_<cprint_action>(arg1))] >> '\n';
			//list = lexeme_d[str_p("list") >> eps_p(anychar_p - (alnum_p | '_'))][list.action = construct_<caction_t>(new_<clist_action>())] >> '\n';

			print = keyword_p("print") >> add_expr[print.action = construct_<caction_t>(new_<cprint_action>(arg1))] >> '\n';
			list = keyword_p("list")[list.action = construct_<caction_t>(new_<clist_action>())] >> '\n';

			statement	= assign[statement.action = arg1]
						| print[statement.action = arg1]
						| list[statement.action = arg1]
						;

			input = *(statement[push_back_a(self.statements_)] 
					| '\n'
					| syntax_error_p);
		}

		rule<ScannerT> const& start() const
		{
			return input;
		}
	};
} ;

// main

int main(int argc, char *argv[])
{
	// եɤ߹ߤˡfile_iteratorѤ
	file_iterator<> first("input.txt");
	if (!first) {
		cout << "ե뤬ץǤޤ" << endl;
		return 1;
	}
	file_iterator<> last = first.make_end();

	vector<caction_t> statements;
	calc_grammer	gr(statements);
	if (parse(first, last, gr, ch_p(' ') | ch_p('\t') | ch_p('\r')).full) {
		// ʸ¹
		for_each(statements.begin(), statements.end(), boost::mem_fn(&caction::action));
	}
	else {
		// 顼äƤΤǡ顼Ȥˤʤ
		// 顼ѡǡ顼Ԥä鼺ԤȤۤ
		// ɤ
		cout << "ʸϼ" << endl;
	}

	return 0;
}
