#include "stdafx.h"
#include <iostream>
#include <iomanip>
#include <memory>
#include "compiler.h"
#include "parser.h"

// 󥹥ȥ饯

compiler::compiler()
	: break_index(-1), error_count(0)
{
}

// ǥȥ饯

compiler::~compiler()
{
}

// ѥ

bool compiler::compile(const std::string &file, vm::data &data)
{
	// ƥॳ
	add_function(vm::SYS_PRINT, TYPE_VOID, "print", "s");
	add_function(vm::SYS_TOSTR, TYPE_STRING, "str", "i");

	// Хѿѡѿơ֥򥻥å
	variables.push_back(CValueTable());
	variables[0].set_global();

	// ƬHALT̿ˤƤ
	OpHalt();

	bool result = script_parser(file, this);	// ʸ

	if (!result)
		return false;							// ѡ顼

	int code_size = LabelSetting();				// ٥˥ɥ쥹
	CraeteData(data, code_size);				// Хʥ
	return error_count == 0;
}

// 顼å

void compiler::error(const std::string& m)
{
	std::cerr << m << std::endl;
	error_count++;
}

// ؿ

bool compiler::add_function(int index, int type, const char *name, const char *args)
{
	CFunctionTag func(type);
	if (!func.SetArgs(args))		// 
		return false;

	func.SetDeclaration();			// Ѥ
	func.SetSystem();				// Systemե饰å
	func.SetIndex(index);			// ƥॳֹ
	if (functions.add(name, func) == 0) {
		return false;
	}
	return true;
}

// ѿ

struct define_value {
	compiler *comp_;
	int type_;
	define_value(compiler *comp, int type): comp_(comp), type_(type)
	{
	}

	void operator()(cnode_t node) const
	{
		comp_->AddValue(type_, node->string(), node->left());
	}
} ;

void compiler::DefineValue(int type, const std::vector<cnode_t> &node)
{
	std::for_each(node.begin(), node.end(), define_value(this, type));
}

// ؿ

void compiler::DefineFunction(int type, const std::string &name, const std::vector<int> &args)
{
	const CFunctionTag *tag = functions.find(name);
	if (tag) {			// Ѥ
		if (!tag->ChkArgList(args)) {
			error("ؿ " + name + " ˰ۤʤ뷿ΰꤵƤޤ");
			return;
		}
	}
	else {
		CFunctionTag func(type);
		func.SetArgs(args);				// 
		func.SetDeclaration();			// Ѥ
		func.SetIndex(MakeLabel());		// ٥Ͽ
		if (functions.add(name, func) == 0) {
			error("顼ؿơ֥ϿǤޤ");
		}
	}
}

// ؿ
//
//	ؿƤФ줿Υå
//
//	+--------------+
//	|     arg2     | -5
//	+--------------+
//	|     arg1     | -4
//	+--------------+
//	|  arg count   | -3
//	+--------------+
//	| base_pointer | -2
//	+--------------+
//	| return addr  | -1
//	+--------------+
//
//	äơγϥɥ쥹-4ȤʤꡢǥȤƤ

// ѿ̾Ͽ
struct add_value {
	compiler *comp_;
	CValueTable &values_;
	mutable int addr_;
	add_value(compiler *comp, CValueTable &values): comp_(comp), values_(values), addr_(-4)
	{
	}

	void operator()(const cargdef &arg) const
	{
		if (!values_.add_arg(arg.type(), arg.name(), addr_)) {
			comp_->error(" " + arg.name() + " ϴϿƤޤ");
		}
		addr_--;
	}
} ;

void compiler::AddFunction(int type, const std::string &name, const std::vector<cargdef> &args, cblock_t block)
{
	CFunctionTag *tag = functions.find(name);
	if (tag) {
		if (tag->IsDefinition()) {
			error("ؿ " + name + " ϴƤޤ");
			return;
		}
		if (tag->IsDeclaration() && !tag->ChkArgList(args)) {
			error("ؿ " + name + " ˰ۤʤ뷿ΰꤵƤޤ");
			return;
		}
		tag->SetDefinition();	// Ѥߤ
	}
	else {
		CFunctionTag func(type);
		func.SetArgs(args);				// 
		func.SetDefinition();			// Ѥ
		func.SetIndex(MakeLabel());		// ٥Ͽ
		tag = functions.add(name, func);
		if (tag == 0)
			error("顼ؿơ֥ϿǤޤ");
	}

	current_function_name = name;		// δؿ̾ϿƤ
	current_function_type = type;		// δؿϿƤ
										// ؿؿҹ¤ˤ̵Τǡ
										// ХѿĤǤ褤

	// ؿΥȥ꡼ݥȤ˥٥֤

	SetLabel(tag->GetIndex());

	BlockIn();		// ѿå䤹

	// ꥹȤϿ
	std::for_each(args.rbegin(), args.rend(), add_value(this, variables.back()));

	// ʸСʸϿ
	if (block) {
		block->analyze(this);
	}

	const CVMCode &code = statement.back();
	if (type == TYPE_VOID) {			// ̵
		if (code.op_ != VM_RETURN)		// return̵ʤ
			OpReturn();					// returnɲ
	}
	else {
		if (code.op_ != VM_RETURNV) {	// return̵ʤ
			error("ؿ " + name + " κǸreturnʸͭޤ");
		}
	}

	BlockOut();		// ѿå򸺤餹

	current_function_name.clear();		// δؿ̾õ
}

// ѿϿ

void compiler::AddValue(int type, const std::string &name, cnode_t node)
{
	int size = 1;
	if (node) {
		if (node->op() != OP_NUMBER) {
			error("ΥǻꤷƤ");
		}
		else if (node->number() <= 0) {
			error("ΥϣʾɬפǤ");
		}
		size = node->number();
	}

	CValueTable &values = variables.back();
	if (!values.add(type, name, size)) {
		error("ѿ " + name + " ϴϿƤޤ");
	}
}

// ٥

int compiler::MakeLabel()
{
	int index = (int)labels.size();
	labels.push_back(CLabel(index));
	return index;
}

// ٥Υߡޥɤ򥹥ơȥȥꥹȤϿ

void compiler::SetLabel(int label)
{
	statement.push_back(CVMCode(VM_MAXCOMMAND, label));
}

// ʸpush

void compiler::PushString(const std::string &str)
{
	PushString((int)text_table.size());
	text_table.insert(text_table.end(), str.begin(), str.end());
	text_table.push_back('\0');
}

// breakʸбJmpޥ

bool compiler::JmpBreakLabel()
{
	if (break_index < 0)
		return false;
	OpJmp(break_index);
	return true;
}

// ֥åǤϡѿåȤѿϿ

void compiler::BlockIn()
{
	int start_addr = 0;					// ѿɥ쥹γϰ
	if (variables.size() > 1) {			// ֥åҤϡϥɥ쥹³ˤ롣
		start_addr = variables.back().size();
	}
	variables.push_back(CValueTable(start_addr));
}

// ֥åνλǡѿפäѿåȤ

void compiler::BlockOut()
{
	variables.pop_back();
}

// ѿѤ˥å

void compiler::AllocStack()
{
	OpAllocStack(variables.back().size());
}

// ٥
//
// ɥ쥹
// ߡΥ٥륳ޥɤͭäɥ쥹򡢥٥ơ֥Ͽ
// Jmpޥɤ٥ơ֥Ͽ줿ɥ쥹ˤ

// ɥ쥹׻
struct calc_addr {
	std::vector<CLabel> &labels_;
	int &pos_;
	calc_addr(std::vector<CLabel> &labels, int &pos): labels_(labels), pos_(pos)
	{
	}
	void operator()(const CVMCode &code)
	{
		if (code.op_ == VM_MAXCOMMAND) {			// ٥Υߡޥ
			labels_[code.arg1_].pos_ = pos_;
		}
		else {
			pos_ += code.size_;
		}
	}
} ;

// ץɥ쥹
struct set_addr {
	std::vector<CLabel> &labels_;
	set_addr(std::vector<CLabel> &labels): labels_(labels)
	{
	}
	void operator()(CVMCode &code)
	{
		switch (code.op_) {
		  case VM_JMP:
		  case VM_JMPC:
		  case VM_JMPNC:
		  case VM_TEST:
		  case VM_CALL:
			code.arg1_ = labels_[code.arg1_].pos_;
			break;
		}
	}
} ;

int compiler::LabelSetting()
{
	// ɥ쥹׻
	int pos = 0;
	std::for_each(statement.begin(), statement.end(), calc_addr(labels, pos));
	// ץɥ쥹
	std::for_each(statement.begin(), statement.end(), set_addr(labels));

	return pos;
}

// Хʥǡ

struct copy_code {
	unsigned char *p;
	copy_code(unsigned char *code): p(code)
	{
	}
	void operator()(const CVMCode &code)
	{
		p = code.Get(p);
	}
} ;

bool compiler::CraeteData(vm::data &data, int code_size)
{
	const CFunctionTag *tag = GetFunctionTag("main");	// ϰ
	if (tag == 0) {
		error("ؿ \"main\" Ĥޤ");
		return false;
	}

	data.command_ = new unsigned char[code_size];
	data.text_buffer_ = new char[text_table.size()];
	data.command_size_ = code_size;
	data.text_size_ = (int)text_table.size();
	data.value_size_ = (int)variables[0].size();
	data.entry_point_ = labels[tag->index_].pos_;

	if (data.text_size_ != 0)
		memcpy(data.text_buffer_, &text_table[0], data.text_size_);

	std::for_each(statement.begin(), statement.end(), copy_code(data.command_));

	return true;
}

// ǥХå
#ifdef	_DEBUG
void compiler::debug_dump()
{
	std::cout << "---variables---" << std::endl;
	size_t vsize = variables.size();
	std::cout << "value stack = " << vsize << std::endl;
	for (size_t i=0; i<vsize; i++) {
		variables[i].dump();
	}
	std::cout << "---code---" << std::endl;

	static const char *op_name[] = {
#define	VM_NAMETABLE
#include "vm_code.h"
#undef	VM_NAMETABLE
		"LABEL",
	} ;

	int	pos = 0;
	size_t size = statement.size();
	for (size_t i=0; i < size; i++) {
		std::cout << std::setw(6) << pos << ": " << op_name[statement[i].op_];
		if (statement[i].size_ > 1) {
			std::cout << ", " << statement[i].arg1_;
		}
		std::cout << std::endl;

		if (statement[i].op_ != VM_MAXCOMMAND) {
			pos += statement[i].size_;
		}
	}
}
#endif
