#include "stdafx.h"
#include <iostream>
#include "node.h"
#include "compiler.h"

// ѿΡɤ
cnode_t cnode::make_node(int op, const std::string &str)
{
	if (op == OP_IDENTIFIER)
		return cnode_t(new cvalue_node(str));

	if (op == OP_STRING) {
		size_t pos = str.rfind('\"');
		if (pos != std::string::npos)
			return cnode_t(new cnode(op, str.substr(0, pos)));
	}
	return cnode_t(new cnode(op, str));
}

// ñ黻ҤΥΡɤ
cnode_t cnode::make_node(int op, cnode_t left)
{
	switch (op) {
	  case OP_NEG:
		if (left->op_ == OP_NUMBER) {			// 黻׻
			left->number_ = -left->number_;
			return left;
		}
		break;
	}
	return cnode_t(new cnode(op, left));
}

// 黻ҤΥΡɤ
cnode_t cnode::make_node(int op, cnode_t left, cnode_t right)
{
	// ΡɤϡleftΡɤleft_Ф˲ä
	if (op == OP_ARRAY) {
		left->left_ = right;
		return left;
	}

	// 黻׻
	if (left->op_ == OP_NUMBER && right->op_ == OP_NUMBER) {
		switch (op) {
		  case OP_LOGAND:
			left->number_ = (left->number_ && right->number_)? 1: 0;
			break;

		  case OP_LOGOR:
			left->number_ = (left->number_ || right->number_)? 1: 0;
			break;

		  case OP_EQ:
			left->number_ = (left->number_ == right->number_)? 1: 0;
			break;

		  case OP_NE:
			left->number_ = (left->number_ != right->number_)? 1: 0;
			break;

		  case OP_GT:
			left->number_ = (left->number_ > right->number_)? 1: 0;
			break;

		  case OP_GE:
			left->number_ = (left->number_ >= right->number_)? 1: 0;
			break;

		  case OP_LT:
			left->number_ = (left->number_ < right->number_)? 1: 0;
			break;

		  case OP_LE:
			left->number_ = (left->number_ <= right->number_)? 1: 0;
			break;

		  case OP_AND:
			left->number_ &= right->number_;
			break;

		  case OP_OR:
			left->number_ |= right->number_;
			break;

		  case OP_LSHIFT:
			left->number_ <<= right->number_;
			break;

		  case OP_RSHIFT:
			left->number_ >>= right->number_;
			break;

		  case OP_SUB:
			left->number_ -= right->number_;
			break;

		  case OP_ADD:
			left->number_ += right->number_;
			break;

		  case OP_MUL:
			left->number_ *= right->number_;
			break;

		  case OP_DIV:
			if (right->number_ == 0) {
				std::cerr << "׻0ǽޤ" << std::endl;
			}
			else {
				left->number_ /= right->number_;
			}
			break;

		  case OP_MOD:
			if (right->number_ == 0) {
				std::cerr << "׻0ǽޤ" << std::endl;
			}
			else {
				left->number_ %= right->number_;
			}
			break;

		  default:
			return cnode_t(new cnode(op, left, right));
		}
		return left;
	}

	// ʸƱΤ׻
	if (left->op_ == OP_STRING && right->op_ == OP_STRING) {
		if (op == OP_ADD) {
			left->string_ += right->string_;
			return left;
		}

		int value = 0;
		switch (op) {
		  case OP_EQ:
			if (left->string_ == right->string_)
				value = 1;
			break;

		  case OP_NE:
			if (left->string_ != right->string_)
				value = 1;
			break;

		  case OP_GT:
			if (left->string_ > right->string_)
				value = 1;
			break;

		  case OP_GE:
			if (left->string_ >= right->string_)
				value = 1;
			break;

		  case OP_LT:
			if (left->string_ < right->string_)
				value = 1;
			break;

		  case OP_LE:
			if (left->string_ <= right->string_)
				value = 1;
			break;

		  default:
			std::cerr << "ʸƱΤǤϤǤʤ׻Ǥ" << std::endl;
			break;
		}
		return cnode_t(new cnode(OP_NUMBER, value));
	}
	return cnode_t(new cnode(op, left, right));
}

// ؿΡɤ
cnode_t cnode::make_node(int op, cnode_t left, cnode_list_t right)
{
	return cnode_t(new cfunction_node(op, left, right));
}

// Ρɤpush
int cnode::push(compiler *c) const
{
	switch (op_) {
	  case OP_NEG:
		if (left_->push(c) == TYPE_STRING)
			c->error("ʸˤϵʤ׻Ǥ");
		c->OpNeg();
		return TYPE_INTEGER;

	  case OP_NUMBER:
		c->PushConst(number_);
		return TYPE_INTEGER;

	  case OP_STRING:
		c->PushString(string_);
		return TYPE_STRING;

	  case OP_FUNCTION:
		return call(c, string_, NULL);
	}

	int left_type = left_->push(c);
	int right_type = right_->push(c);

	if (left_type != right_type)
		c->error("ʸ֤Ƿ׻Ǥޤ");

	// ׻Ρɤν
	if (left_type == TYPE_INTEGER) {
		switch (op_) {
		  case OP_LOGAND:
			c->OpLogAnd();
			break;

		  case OP_LOGOR:
			c->OpLogOr();
			break;

		  case OP_EQ:
			c->OpEq();
			break;

		  case OP_NE:
			c->OpNe();
			break;

		  case OP_GT:
			c->OpGt();
			break;

		  case OP_GE:
			c->OpGe();
			break;

		  case OP_LT:
			c->OpLt();
			break;

		  case OP_LE:
			c->OpLe();
			break;

		  case OP_AND:
			c->OpAnd();
			break;

		  case OP_OR:
			c->OpOr();
			break;

		  case OP_LSHIFT:
			c->OpLeftShift();
			break;

		  case OP_RSHIFT:
			c->OpRightShift();
			break;

		  case OP_SUB:
			c->OpSub();
			break;

		  case OP_ADD:
			c->OpAdd();
			break;

		  case OP_MUL:
			c->OpMul();
			break;

		  case OP_DIV:
			c->OpDiv();
			break;

		  case OP_MOD:
			c->OpMod();
			break;

		  default:
			c->error("顼Ǥʤ׻Ρɤޤ");
			break;
		}
		return TYPE_INTEGER;
	}

	// ʸ׻Ρɤν
	switch (op_) {
	  case OP_EQ:
		c->OpStrEq();
		return TYPE_INTEGER;

	  case OP_NE:
		c->OpStrNe();
		return TYPE_INTEGER;

	  case OP_GT:
		c->OpStrGt();
		return TYPE_INTEGER;

	  case OP_GE:
		c->OpStrGe();
		return TYPE_INTEGER;

	  case OP_LT:
		c->OpStrLt();
		return TYPE_INTEGER;

	  case OP_LE:
		c->OpStrLe();
		return TYPE_INTEGER;

	  case OP_ADD:
		c->OpStrAdd();
		break;

	  default:
		c->error("ʸǤϷ׻ǤʤǤ");
		break;
	}
	return TYPE_STRING;
}

// Ρɤpop
// ׻ΡɤpopǤʤ

int cnode::pop(compiler *c) const
{
	c->error("顼׻ΡɤpopƤޤ");
	return TYPE_INTEGER;
}

// ʸ

void cnode::assign(compiler *c) const
{
	if (op_ != OP_ASSIGN)
		left_->push(c);

	if (right_->push(c) == TYPE_INTEGER) {
		switch (op_) {
		  case OP_ADD_ASSIGN:
			c->OpAdd();
			break;

		  case OP_SUB_ASSIGN:
			c->OpSub();
			break;

		  case OP_MUL_ASSIGN:
			c->OpMul();
			break;

		  case OP_DIV_ASSIGN:
			c->OpDiv();
			break;

		  case OP_MOD_ASSIGN:
			c->OpMod();
			break;

		  case OP_AND_ASSIGN:
			c->OpAnd();
			break;

		  case OP_OR_ASSIGN:
			c->OpOr();
			break;

		  case OP_LSHIFT_ASSIGN:
			c->OpLeftShift();
			break;

		  case OP_RSHIFT_ASSIGN:
			c->OpRightShift();
			break;
		}
		if (left_->pop(c) != TYPE_INTEGER)
			c->error("ʸ󷿤Ƥޤ");
		return;
	}

	switch (op_) {
	  case OP_ADD_ASSIGN:
		c->OpStrAdd();
		break;

	  case OP_ASSIGN:
		break;

	  default:
		c->error("ʸǤϵʤ׻Ǥ");
		break;
	}
	if (left_->pop(c) != TYPE_STRING)
		c->error("ʸƤޤ");
}

// ѿΡɤpush
int cvalue_node::push(compiler *c) const
{
	if (op_ != OP_IDENTIFIER) {
		c->error("顼ѿΡɤѿʳϿƤޤ");
	}
	else {
		const CValueTag *tag = c->GetValueTag(string_);
		if (tag == 0) {
			c->error("ѿ " + string_ + " Ƥޤ");
		}
		else {
			// ȷѿϡˤ¸ߤʤ
			if ((tag->type_ & TYPE_REF) != 0) {
				if (left_) {		// 
					left_->push(c);
					c->PushLocalArrayRef(tag->addr_);
				}
				else {
					c->PushLocalRef(tag->addr_);
				}
				return tag->type_ & ~TYPE_REF;
			}
			if (tag->global_) {		// ѿ
				if (left_) {		// 
					left_->push(c);
					c->PushArray(tag->addr_);
				}
				else {
					c->PushValue(tag->addr_);
				}
			}
			else {					// ѿ
				if (left_) {		// 
					left_->push(c);
					c->PushLocalArray(tag->addr_);
				}
				else {
					c->PushLocal(tag->addr_);
				}
			}
			return tag->type_;
		}
	}
	return TYPE_INTEGER;
}

// ѿΡɤpop
int cvalue_node::pop(compiler *c) const
{
	if (op_ != OP_IDENTIFIER) {
		c->error("顼ѿΡɤѿʳϿƤޤ");
	}
	else {
		const CValueTag *tag = c->GetValueTag(string_);
		if (tag == 0) {
			c->error("ѿ " + string_ + " Ƥޤ");
		}
		else {
			// ȷѿϡˤ¸ߤʤ
			if ((tag->type_ & TYPE_REF) != 0) {
				if (left_) {		// 
					left_->push(c);
					c->PopLocalArrayRef(tag->addr_);
				}
				else {
					c->PopLocalRef(tag->addr_);
				}
				return tag->type_ & ~TYPE_REF;
			}
			if (tag->global_) {		// ѿ
				if (left_) {		// 
					left_->push(c);
					c->PopArray(tag->addr_);
				}
				else {
					c->PopValue(tag->addr_);
				}
			}
			else {					// ѿ
				if (left_) {		// 
					left_->push(c);
					c->PopLocalArray(tag->addr_);
				}
				else {
					c->PopLocal(tag->addr_);
				}
			}
			return tag->type_;
		}
	}
	return TYPE_INTEGER;
}

// ؿƤӽФ
struct set_arg {
	compiler *comp_;
	const CFunctionTag *func_;
	mutable int index_;
	set_arg(compiler *comp, const CFunctionTag *func): comp_(comp), func_(func), index_(0)
	{
	}

	void operator()(cnode_t node) const
	{
		int type = func_->GetArg(index_++);
		if ((type & TYPE_REF) != 0) {		// 
			if (node->op() != OP_IDENTIFIER) {
				comp_->error("ȷˡѿʳϻǤޤ");
			}
			else {
				const CValueTag *tag = comp_->GetValueTag(node->string());
				if (tag == 0) {
					comp_->error("ѿ " + node->string() + " Ƥޤ");
				}
				else if (tag->type_ >= TYPE_INTEGER_REF) {		// 
					// ȷѿϡ뤷ʤ
					if (node->left()) {
						node->left()->push(comp_);
						comp_->PushLocal(tag->addr_);
						comp_->OpAdd();
					}
					else {
						comp_->PushLocal(tag->addr_);
					}
				}
				else {
					if ((tag->type_ | TYPE_REF) != type) {
						comp_->error("η礤ޤ");
					}
					int addr = tag->addr_;
					if (tag->global_)			// ѿ
						addr |= vm::vcpu::global_flag;
					// ɥ쥹push
					if (node->left()) {			// 
						if (node->left()->op() == OP_NUMBER) {
							comp_->PushAddr(addr + node->left()->number());
						}
						else {
							node->left()->push(comp_);
							comp_->PushArrayAddr(addr);
						}
					}
					else {
						comp_->PushAddr(addr);
					}
				}
			}
		}
		else {
			if (node->push(comp_) != type) {
				comp_->error("η礤ޤ");
			}
		}
	}
} ;

// ؿƤӽФ
int cnode::call(compiler *c, const std::string &name, const std::vector<cnode_t> *args) const
{
	const CFunctionTag *tag = c->GetFunctionTag(name);
	if (tag == NULL) {
		c->error("ؿ " + name + "ϡƤޤ");
		return TYPE_INTEGER;
	}

	int arg_size = (args)? (int)args->size(): 0;
	if (tag->ArgSize() != arg_size) {
		c->error("ο礤ޤ");
	}

	// push
	if (args && tag->ArgSize() == arg_size) {
		std::for_each(args->begin(), args->end(), set_arg(c, tag));
	}

	// οpush
	c->PushConst(arg_size);

	if (tag->IsSystem()) {
		c->OpSysCall(tag->GetIndex());		// ƥॳ
	}
	else {
		c->OpCall(tag->GetIndex());			// ץȾδؿ
	}

	return tag->type_;
}

int cfunction_node::push(compiler *c) const
{
	return call(c, left_->string(), &node_list_->args_);
}

// ؿpopϤǤʤΤǥ顼åФ
int cfunction_node::pop(compiler *c) const
{
	c->error("顼ؿΡɤpop");
	return TYPE_INTEGER;
}

// ʸΡ
cstatement_t cstatement::make_statement(int state)
{
	switch (state) {
	case NOP_STATE:
		return cstatement_t(new cnop_statement());

	case DEFAULT_STATE:
		return cstatement_t(new cdefault_statement());

	case BREAK_STATE:
		return cstatement_t(new cbreak_statement());

	case RETURN_STATE:
		return cstatement_t(new creturn_statement());

	case IF_STATE:
		return cstatement_t(new cif_statement());

	case FOR_STATE:
		return cstatement_t(new cfor_statement());

	case WHILE_STATE:
		return cstatement_t(new cwhile_statement());
	}

	std::cerr << "顼ʸΡɥߥ" << std::endl;
	return cstatement_t(new cnop_statement());
}

cstatement_t cstatement::make_statement(int state, cnode_t node)
{
	switch (state) {
	case ASSIGN_STATE:
		return cstatement_t(new cassign_statement(node));

	case CASE_STATE:
		return cstatement_t(new ccase_statement(node));

	case SWITCH_STATE:
		return cstatement_t(new cswitch_statement(node));

	case CALL_STATE:
		return cstatement_t(new ccall_statement(node));
	}

	std::cerr << "顼ʸΡɥߥ" << std::endl;
	return cstatement_t(new cnop_statement());
}

cstatement_t cstatement::make_statement(int state, cblock_t block)
{
	switch (state) {
	case BLOCK_STATE:
		return cstatement_t(new cblock_statement(block));
	}

	std::cerr << "顼ʸΡɥߥ" << std::endl;
	return cstatement_t(new cnop_statement());
}

// nopʸ
void cnop_statement::analyze(compiler *c) const
{
}

// ʸ
void cassign_statement::analyze(compiler *c) const
{
	node_->assign(c);
}

// ؿƤӽФʸ
void ccall_statement::analyze(compiler *c) const
{
	int type = node_->push(c);
	if (type != TYPE_VOID)
		c->OpPop();			// ͤΤƤ뤿pop
}

// caseʸ
void ccase_statement::analyze(compiler *c) const
{
	c->SetLabel(label_);
}

// caseʸ
void ccase_statement::case_analyze(compiler *c, int *default_label)
{
	label_ = c->MakeLabel();
	if (node_->op() != OP_NUMBER)
		c->error("case ʸˤΤ߻Ǥޤ");
	node_->push(c);
	c->OpTest(label_);
}

// defaultʸ
void cdefault_statement::analyze(compiler *c) const
{
	c->SetLabel(label_);
}

// defaultʸ
void cdefault_statement::case_analyze(compiler *c, int *default_label)
{
	label_ = c->MakeLabel();
	*default_label = label_;
}

// breakʸ
void cbreak_statement::analyze(compiler *c) const
{
	if (!c->JmpBreakLabel()) {
		c->error("breakswitch/for/whileͭޤ");
	}
}

// returnʸ
void creturn_statement::analyze(compiler *c) const
{
	if (c->GetFunctionType() == TYPE_VOID) {	// ̵
		if (node_ != 0) {
			c->error("voidؿͤꤵƤޤ");
		}
		c->OpReturn();
	}
	else {
		if (node_ == 0) {
			c->error("ؿͤޤ");
		}
		else {
			int node_type = node_->push(c);		// ͤpush
			if (node_type != c->GetFunctionType()) {
				c->error("ͤη礤ޤ");
			}
		}
		c->OpReturnV();
	}
}

// ifʸ
void cif_statement::analyze(compiler *c) const
{
	node_->push(c);
	int label1 = c->MakeLabel();
	c->OpJmpNC(label1);
	statement_[0]->analyze(c);

	if (statement_[1]) {
		int label2 = c->MakeLabel();
		c->OpJmp(label2);
		c->SetLabel(label1);
		statement_[1]->analyze(c);
		c->SetLabel(label2);
	}
	else {
		c->SetLabel(label1);
	}
}

// forʸ
void cfor_statement::analyze(compiler *c) const
{
	int label1 = c->MakeLabel();
	int label2 = c->MakeLabel();

	int break_label = c->SetBreakLabel(label2);

	node_[0]->assign(c);
	c->SetLabel(label1);
	node_[1]->push(c);
	c->OpJmpNC(label2);
	statement_->analyze(c);
	node_[2]->assign(c);
	c->OpJmp(label1);
	c->SetLabel(label2);

	c->SetBreakLabel(break_label);
}

// whileʸ
void cwhile_statement::analyze(compiler *c) const
{
	int label1 = c->MakeLabel();
	int label2 = c->MakeLabel();

	int break_label = c->SetBreakLabel(label2);

	c->SetLabel(label1);
	node_->push(c);
	c->OpJmpNC(label2);
	statement_->analyze(c);
	c->OpJmp(label1);
	c->SetLabel(label2);

	c->SetBreakLabel(break_label);
}

// switchʸ
void cswitch_statement::analyze(compiler *c) const
{
	if (!statement_.empty()) {
		node_->push(c);

		int label = c->MakeLabel();		// L0٥
		int break_label = c->SetBreakLabel(label);
		int default_label = label;

		std::for_each(statement_.begin(), statement_.end(),
			boost::bind(&cstatement::case_analyze, _1, c, &default_label));

		c->OpPop();
		c->OpJmp(default_label);

		std::for_each(statement_.begin(), statement_.end(), boost::bind(&cstatement::analyze, _1, c));
		c->SetLabel(label);

		c->SetBreakLabel(break_label);
	}
}

// blockʸ
void cblock_statement::analyze(compiler *c) const
{
	c->BlockIn();
	block_->analyze(c);
	c->BlockOut();
}

// ʸ֥å
void cblock::analyze(compiler *c) const
{
	std::for_each(decl_.begin(), decl_.end(), boost::bind(&cdeclaration::analyze, _1, c));
	if (!decl_.empty())
		c->AllocStack();	// åե졼
	std::for_each(state_.begin(), state_.end(), boost::bind(&cstatement::analyze, _1, c));
}

// β
void cdeclaration::analyze(compiler *c) const
{
	if (is_func_) {		// ؿ
		c->DefineFunction(type_, name_, arg_);
	}
	else {
		c->DefineValue(type_, node_);
	}
}

// ؿβ
void cfunction::analyze(compiler *c) const
{
	c->AddFunction(type_, name_, args_, block_);
}
