#include "stdafx.h"
#include <iostream>
#include <iomanip>
#include <memory>
#include "compiler.h"
#include "parser.h"

// RXgN^

compiler::compiler()
	: break_index(-1), error_count(0)
{
}

// fXgN^

compiler::~compiler()
{
}

// RpC

bool compiler::compile(const std::string &file, vm::data &data)
{
	// VXeR[̐ݒ
	add_function(vm::SYS_PRINT, TYPE_VOID, "print", "s");
	add_function(vm::SYS_TOSTR, TYPE_STRING, "str", "i");

	// O[oϐpAϐe[uZbg
	variables.push_back(CValueTable());
	variables[0].set_global();

	// 擪HALT߂ɂĂ
	OpHalt();

	bool result = script_parser(file, this);	// \

	if (!result)
		return false;							// p[T[G[

	int code_size = LabelSetting();				// xɃAhXݒ
	CraeteData(data, code_size);				// oCi
	return error_count == 0;
}

// G[bZ[Wo

void compiler::error(const std::string& m)
{
	std::cerr << m << std::endl;
	error_count++;
}

// ֐̒`

bool compiler::add_function(int index, int type, const char *name, const char *args)
{
	CFunctionTag func(type);
	if (!func.SetArgs(args))		// ݒ
		return false;

	func.SetDeclaration();			// 錾ς
	func.SetSystem();				// SystemtOZbg
	func.SetIndex(index);			// VXeR[ԍݒ
	if (functions.add(name, func) == 0) {
		return false;
	}
	return true;
}

// Oϐ̒`

struct define_value {
	compiler *comp_;
	int type_;
	define_value(compiler *comp, int type): comp_(comp), type_(type)
	{
	}

	void operator()(cnode_t node) const
	{
		comp_->AddValue(type_, node->string(), node->left());
	}
} ;

void compiler::DefineValue(int type, const std::vector<cnode_t> &node)
{
	std::for_each(node.begin(), node.end(), define_value(this, type));
}

// ֐錾

void compiler::DefineFunction(int type, const std::string &name, const std::vector<int> &args)
{
	const CFunctionTag *tag = functions.find(name);
	if (tag) {			// ɐ錾ς
		if (!tag->ChkArgList(args)) {
			error("֐ " + name + " ɈقȂ^̈w肳Ă܂");
			return;
		}
	}
	else {
		CFunctionTag func(type);
		func.SetArgs(args);				// ݒ
		func.SetDeclaration();			// 錾ς
		func.SetIndex(MakeLabel());		// xo^
		if (functions.add(name, func) == 0) {
			error("G[F֐e[uɓo^ł܂");
		}
	}
}

// ֐`
//
//	֐Ă΂ꂽ_̃X^bN
//
//	+--------------+
//	|     arg2     | -5
//	+--------------+
//	|     arg1     | -4
//	+--------------+
//	|  arg count   | -3
//	+--------------+
//	| base_pointer | -2
//	+--------------+
//	| return addr  | -1
//	+--------------+
//
//	āÅJnAhX-4ƂȂAfNgĂB

// ̕ϐo^
struct add_value {
	compiler *comp_;
	CValueTable &values_;
	mutable int addr_;
	add_value(compiler *comp, CValueTable &values): comp_(comp), values_(values), addr_(-4)
	{
	}

	void operator()(const cargdef &arg) const
	{
		if (!values_.add_arg(arg.type(), arg.name(), addr_)) {
			comp_->error(" " + arg.name() + " ͊ɓo^Ă܂B");
		}
		addr_--;
	}
} ;

void compiler::AddFunction(int type, const std::string &name, const std::vector<cargdef> &args, cblock_t block)
{
	CFunctionTag *tag = functions.find(name);
	if (tag) {
		if (tag->IsDefinition()) {
			error("֐ " + name + " ͊ɒ`Ă܂");
			return;
		}
		if (tag->IsDeclaration() && !tag->ChkArgList(args)) {
			error("֐ " + name + " ɈقȂ^̈w肳Ă܂");
			return;
		}
		tag->SetDefinition();	// `ς݂ɐݒ
	}
	else {
		CFunctionTag func(type);
		func.SetArgs(args);				// ݒ
		func.SetDefinition();			// `ς
		func.SetIndex(MakeLabel());		// xo^
		tag = functions.add(name, func);
		if (tag == 0)
			error("G[F֐e[uɓo^ł܂");
	}

	current_function_name = name;		// ̊֐o^Ă
	current_function_type = type;		// ̊֐^o^Ă
										// ֐֐iq\ĵ͖ŁA
										// O[oϐPł悢

	// ֐̃Gg[|CgɃxu

	SetLabel(tag->GetIndex());

	BlockIn();		// ϐX^bN𑝂₷

	// Xgo^
	std::for_each(args.rbegin(), args.rend(), add_value(this, variables.back()));

	// ΁Ao^
	if (block) {
		block->analyze(this);
	}

	const CVMCode &code = statement.back();
	if (type == TYPE_VOID) {			// ߂l
		if (code.op_ != VM_RETURN)		// returnȂ
			OpReturn();					// returnǉ
	}
	else {
		if (code.op_ != VM_RETURNV) {	// returnȂ
			error("֐ " + name + " ̍ŌreturnL܂B");
		}
	}

	BlockOut();		// ϐX^bN炷

	current_function_name.clear();		// ̊֐
}

// ϐ̓o^

void compiler::AddValue(int type, const std::string &name, cnode_t node)
{
	int size = 1;
	if (node) {
		if (node->op() != OP_NUMBER) {
			error("z̃TCY͒萔Ŏw肵ĂB");
		}
		else if (node->number() <= 0) {
			error("z̃TCY͂Pȏ̒萔KvłB");
		}
		size = node->number();
	}

	CValueTable &values = variables.back();
	if (!values.add(type, name, size)) {
		error("ϐ " + name + " ͊ɓo^Ă܂B");
	}
}

// x

int compiler::MakeLabel()
{
	int index = (int)labels.size();
	labels.push_back(CLabel(index));
	return index;
}

// x̃_~[R}hXe[ggXgɓo^

void compiler::SetLabel(int label)
{
	statement.push_back(CVMCode(VM_MAXCOMMAND, label));
}

// 萔push

void compiler::PushString(const std::string &str)
{
	PushString((int)text_table.size());
	text_table.insert(text_table.end(), str.begin(), str.end());
	text_table.push_back('\0');
}

// breakɑΉJmpR}h

bool compiler::JmpBreakLabel()
{
	if (break_index < 0)
		return false;
	OpJmp(break_index);
	return true;
}

// ubNł́AVϐZbgɕϐo^

void compiler::BlockIn()
{
	int start_addr = 0;					// ϐAhX̊Jnʒu
	if (variables.size() > 1) {			// ubN̓q́AJnAhX𑱂ɂB
		start_addr = variables.back().size();
	}
	variables.push_back(CValueTable(start_addr));
}

// ubN̏IŁAϐXR[viϐZbg폜j

void compiler::BlockOut()
{
	variables.pop_back();
}

// [JϐpɃX^bNm

void compiler::AllocStack()
{
	OpAllocStack(variables.back().size());
}

// x
//
// PDAhX𐶐
// QD_~[̃xR}hLAhXAxe[uɓo^
// RDJmpR}h̔ѐxe[uɓo^ꂽAhXɂ

// AhXvZ
struct calc_addr {
	std::vector<CLabel> &labels_;
	int &pos_;
	calc_addr(std::vector<CLabel> &labels, int &pos): labels_(labels), pos_(pos)
	{
	}
	void operator()(const CVMCode &code)
	{
		if (code.op_ == VM_MAXCOMMAND) {			// x̃_~[R}h
			labels_[code.arg1_].pos_ = pos_;
		}
		else {
			pos_ += code.size_;
		}
	}
} ;

// WvAhXݒ
struct set_addr {
	std::vector<CLabel> &labels_;
	set_addr(std::vector<CLabel> &labels): labels_(labels)
	{
	}
	void operator()(CVMCode &code)
	{
		switch (code.op_) {
		  case VM_JMP:
		  case VM_JMPC:
		  case VM_JMPNC:
		  case VM_TEST:
		  case VM_CALL:
			code.arg1_ = labels_[code.arg1_].pos_;
			break;
		}
	}
} ;

int compiler::LabelSetting()
{
	// AhXvZ
	int pos = 0;
	std::for_each(statement.begin(), statement.end(), calc_addr(labels, pos));
	// WvAhXݒ
	std::for_each(statement.begin(), statement.end(), set_addr(labels));

	return pos;
}

// oCif[^

struct copy_code {
	unsigned char *p;
	copy_code(unsigned char *code): p(code)
	{
	}
	void operator()(const CVMCode &code)
	{
		p = code.Get(p);
	}
} ;

bool compiler::CraeteData(vm::data &data, int code_size)
{
	const CFunctionTag *tag = GetFunctionTag("main");	// Jnʒu
	if (tag == 0) {
		error("֐ \"main\" ܂B");
		return false;
	}

	data.command_ = new unsigned char[code_size];
	data.text_buffer_ = new char[text_table.size()];
	data.command_size_ = code_size;
	data.text_size_ = (int)text_table.size();
	data.value_size_ = (int)variables[0].size();
	data.entry_point_ = labels[tag->index_].pos_;

	if (data.text_size_ != 0)
		memcpy(data.text_buffer_, &text_table[0], data.text_size_);

	std::for_each(statement.begin(), statement.end(), copy_code(data.command_));

	return true;
}

// fobO_v
#ifdef	_DEBUG
void compiler::debug_dump()
{
	std::cout << "---variables---" << std::endl;
	size_t vsize = variables.size();
	std::cout << "value stack = " << vsize << std::endl;
	for (size_t i=0; i<vsize; i++) {
		variables[i].dump();
	}
	std::cout << "---code---" << std::endl;

	static const char *op_name[] = {
#define	VM_NAMETABLE
#include "vm_code.h"
#undef	VM_NAMETABLE
		"LABEL",
	} ;

	int	pos = 0;
	size_t size = statement.size();
	for (size_t i=0; i < size; i++) {
		std::cout << std::setw(6) << pos << ": " << op_name[statement[i].op_];
		if (statement[i].size_ > 1) {
			std::cout << ", " << statement[i].arg1_;
		}
		std::cout << std::endl;

		if (statement[i].op_ != VM_MAXCOMMAND) {
			pos += statement[i].size_;
		}
	}
}
#endif
